#include "IZ_midi.h"

char* IZ_GetMIDINoteName(uint8_t midi_note) {
	static const char* pitch_names[] = {
		"C",
		"C#",
		"D",
		"D#",
		"E",
		"F",
		"F#",
		"G",
		"G#",
		"A",
		"A#",
		"B"
	};

	const uint8_t pitch_class = midi_note % 12;
	const uint8_t octave = midi_note / 12;
	static char note_name[4];
	sprintf_s(note_name, 4, "%s%u", pitch_names[pitch_class], octave);
	return note_name;
}

uint8_t IZ_GetMIDINoteFromName(char* name) {
	char name_copy[4];
	memcpy_s(name_copy, 4, name, 4);
	_strlwr_s(name_copy, 4);

	uint8_t octave;
	const char base_pitch_name[] = "c d ef g a b";

	if (strlen(name_copy) == 2) {
		octave = name_copy[1] - '0';
		for (uint8_t i = 0; i < 12; i += 1) {
			if (base_pitch_name[i] == name_copy[0]) {
				return (octave * 12) + i;
			}
		}
		return 255u;
	}

	uint8_t pitch_class;
	octave = name_copy[2] - '0';
	if (strstr(name_copy, "c#") || strstr(name_copy, "db")) {
		pitch_class = 1;
	} else if (strstr(name_copy, "d#") || strstr(name_copy, "eb")) {
		pitch_class = 3;
	} else if (strstr(name_copy, "f#") || strstr(name_copy, "gb")) {
		pitch_class = 6;
	} else if (strstr(name_copy, "g#") || strstr(name_copy, "ab")) {
		pitch_class = 8;
	} else if (strstr(name_copy, "a#") || strstr(name_copy, "bb")) {
		pitch_class = 10;
	} else {
		return 255u;
	}

	return (octave * 12) + pitch_class;
}

void IZ_HandleMIDINoteOnOffEvents(PmEvent e, IZ_MIDIInputState* state, IZ_Action* action) {
	uint32_t message = e.message;
	uint8_t status = message & 0xF0u;
	uint8_t channel = message & 0x0Fu;
	uint8_t data1 = (message >> 8) & 0xFFu;
	// uint8_t data2 = (message >> 16) & 0xFFu;

	for (uint8_t i = 0; i < CONTROLS; i += 1) {
		if (
			data1 == state->config.control_mapping[i]
			&& (
				(state->config.channel < 16 && channel == state->config.channel)
				|| state->config.channel >= 16
			)
		) {
			const uint16_t bitflag = (0x1 << i);
			if (status == IZ_MIDI_NOTE_ON) {
				*action |= bitflag;
				return;
			}
			if (status == IZ_MIDI_NOTE_OFF) {
				*action &= ~bitflag;
				return;
			}
		}
	}
}

void IZ_HandleMIDIInputEvents(PmEvent e, IZ_MIDIInputState(* state)[PLAYERS], IZ_Action(* action)[PLAYERS]) {
	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		IZ_HandleMIDINoteOnOffEvents(e, state[player_index], action[player_index]);
	}
}

IZ_ProcedureResult IZ_SaveMIDIInputConfig(const char* config_path, IZ_MIDIInputState(* state)[PLAYERS]) {
	uint8_t problem = 0;

	char control_mapping_section_name[27];
	char main_section_name[12];

	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		sprintf_s(control_mapping_section_name, 27, "MIDIInput.%d.ControlMapping", player_index);
		for (uint8_t i = 0; i < CONTROLS; i += 1) {
			if (!ini_puts(
				control_mapping_section_name,
				ACTION_NAMES[i],
				IZ_GetMIDINoteName(state[player_index]->config.control_mapping[i]),
				config_path
			)) {
				problem |= (1 << player_index);
			}
		}

		sprintf_s(main_section_name, 12, "MIDIInput.%d", player_index);
		if (!ini_putl(
			main_section_name,
			"Channel",
			state[player_index]->config.channel,
			config_path
		)) {
			problem |= (1 << player_index);
		}

		if (!ini_putl(
			main_section_name,
			"DeviceID",
			state[player_index]->config.device_id,
			config_path
		)) {
			problem |= (1 << player_index);
		}
	}

	return problem;
}

void IZ_LoadMIDIInputConfig(const char* config_path, IZ_MIDIInputState(* state)[PLAYERS]) {
	char buffer[128];
	char control_mapping_section_name[27];
	char main_section_name[12];

	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		sprintf_s(control_mapping_section_name, 27, "MIDIInput.%d.ControlMapping", player_index);
		for (uint8_t i = 0; i < CONTROLS; i += 1) {
			ini_gets(
				control_mapping_section_name,
				ACTION_NAMES[i],
				IZ_GetMIDINoteName(IZ_DEFAULT_MIDI_INPUT_STATE[player_index].config.control_mapping[i]),
				buffer,
				128,
				config_path
			);

			state[player_index]->config.control_mapping[i] = IZ_GetMIDINoteFromName(buffer);
		}

		sprintf_s(main_section_name, 12, "MIDIInput.%d", player_index);
		state[player_index]->config.channel = ini_getl(main_section_name, "Channel", player_index, config_path);
		state[player_index]->config.device_id = ini_getl(main_section_name, "DeviceID", player_index, config_path);
	}
}

IZ_ProcedureResult IZ_InitializeMIDIInput(const char* config_path, IZ_MIDIInputState(* state)[PLAYERS]) {
	if (Pm_Initialize()) {
		return 1;
	}

	SDL_memcpy(state, &IZ_DEFAULT_MIDI_INPUT_STATE, sizeof(IZ_MIDIInputState));
	IZ_LoadMIDIInputConfig(config_path, state);
	if (IZ_SaveMIDIInputConfig(config_path, state)) {
		return 2;
	}

	for (uint8_t player_index = 0; player_index < PLAYERS; player_index += 1) {
		state[player_index]->device_info = Pm_GetDeviceInfo(state[player_index]->config.device_id);
		state[player_index]->stream = NULL;
		Pm_OpenInput(
			&state[player_index]->stream,
			state[player_index]->config.device_id,
			NULL,
			MIDI_EVENT_BUFFER_SIZE,
			NULL,
			NULL
		);
	}

	return 0;
}

void IZ_TeardownMIDIInput(IZ_MIDIInputState(* state)[PLAYERS]) {
	for (uint8_t i = 0; i < PLAYERS; i += 1) {
		if (!state[i]->stream) {
			continue;
		}
		Pm_Close(state[i]->stream);
	}
}
